from typing import Literal, Union, Optional, List
from .upload_manager import UploadManager
from tqdm.notebook import tqdm
from .logger import log
import requests
import ipyplot
import json
import time


class manotAI:

    def __init__(self, url: str, token: str) -> None:
        self.__url = url
        self.__token = token

    def setup(
            self,
            name: str,
            images_path: str,
            ground_truths_path: str,
            detections_path: str,
            detections_metadata_format: Literal['cxcywh', 'xywh', 'xyx2y2'],
            classes_txt_path: str,
            data_provider: Literal['s3', 'local']
    ) -> Union[bool, dict]:

        url = f"{self.__url}/api/v1/setup/"
        data = {
            "name": name,
            "images_path": images_path,
            "ground_truths_path": ground_truths_path,
            "detections_path": detections_path,
            "detections_metadata_format": detections_metadata_format,
            "classes_txt_path": classes_txt_path,
            "data_provider": data_provider
        }

        try:
            response = requests.post(url=url, data=json.dumps(data), headers={"token": self.__token})
        except Exception:
            log.error("There is problem with request.")
            return False

        if response.status_code == 202:
            log.info("Setup is being prepared to be started.")
            if self.__check_progress(self.get_setup, response.json()["id"]):
                log.info("Setup has successfully created.")
            else:
                log.error(f'There is problem setup process with id {response.json()["id"]}.')

            return response.json()

        log.error(response.text)
        return False

    def get_setup(self, setup_id: int) -> Union[bool, None, dict]:

        url = f"{self.__url}/api/v1/setup/{setup_id}"

        try:
            response = requests.get(url=url)
        except Exception:
            log.error("There is problem with request.")
            return False

        if response.status_code != 200:
            log.warning("Setup not found.")
            return None

        return response.json()

    def insight(
            self,
            name: str,
            setup_id: int,
            data_path: Union[str, List[str]],
            data_provider: Literal['s3', 'local', 'deeplake'],
            weight_name: Optional[str] = None,
    ) -> Union[bool, dict]:

        url = f"{self.__url}/api/v1/insight/"
        data = {
            "name": name,
            "setup_id": setup_id,
            "data_path": data_path,
            "data_provider": data_provider
        }
        if weight_name:
            data["weight_name"] = weight_name

        try:
            response = requests.post(url=url, data=json.dumps(data), headers={"token": self.__token})
        except Exception:
            log.error("There is problem with request.")
            return False

        if response.status_code == 202:
            log.info("Insights process is successfully started.")
            return response.json()

        log.error(response.text)
        return False

    def get_insight(self, insight_id: int) -> Union[bool, None, dict]:

        url = f"{self.__url}/api/v1/insight/{insight_id}"

        try:
            response = requests.get(url=url)
        except Exception:
            log.error("There is problem with request.")
            return False

        if response.status_code != 200:
            log.warning("Insight not found.")
            return None

        log.info("Insight is successfully found.")
        return response.json()


    def visualize_data_set(self, data_set_id: int):
        url = f"{self.__url}/api/v1/data_set/{data_set_id}"

        try:
            response = requests.get(url=url).json()
        except Exception:
            log.error("There is problem with request.")
            return False

        if not response:
            log.warning("Data Set not found.")
            return None

        images = response['data_set_images']
        images_urls = []
        for file in images:
            images_urls.append(self.__process(file['id']))

        return ipyplot.plot_images(images_urls, img_width=200, show_url=False)


    def upload_data(self, dir_path: str, process: Literal["setup", "insight"]):

        upload_manager = UploadManager(directory=dir_path, url=self.__url, token=self.__token)
        if process == "setup":
            return upload_manager.upload_setup_data()
        elif process == "insight":
            return upload_manager.upload_insight_data()
        else:
            log.error("Process must be 'setup' or 'insight'.")
            return False


    def __process(self, image_id: int):
        url = f"{self.__url}/api/v1/image/{image_id}"
        return url


    def __check_progress(self, process_method, id):
        progress = 0
        progress_bar = tqdm(desc="Progress", total=100)
        while progress < 100:
            result = process_method(id)
            if result and result["status"] != "failure":
                progress_bar.update(result['progress'] - progress)
                progress = result['progress']
            else:
                progress_bar.close()
                return False
            time.sleep(2)
        progress_bar.close()
        return True




from tqdm.notebook import tqdm
import concurrent.futures
from .logger import log
import requests
import os.path
import glob
import uuid


class UploadManager:

    def __init__(self, directory: str, url: str, token: str):
        self.directory = directory
        self.url = url
        self.token = token
        self.save_path = str(uuid.uuid4())


    def upload_insight_data(self):

        if not self._is_exist(self.directory):
            log.error("No such directory.")
            return False

        correct_files = []
        for file_path in glob.glob(self.directory + "/*"):
            if file_path.endswith((".jpeg", ".jpg", ".png", ".avi", ".gif", ".m4v", ".mkv", ".mp4")):
                correct_files.append({"type": "image", "path": file_path})

        if not correct_files:
            log.error("There are no files to upload.")
            return False

        if self.upload_data(correct_files):
            return {"data_path": os.path.join(self.save_path, "images")}


    def upload_setup_data(self):

        images_path = os.path.join(self.directory, "images")
        detections_path = os.path.join(self.directory, "detections")
        ground_truths_path = os.path.join(self.directory, "ground_truths")
        classes_txt_path = os.path.join(self.directory, "classes.txt")
        if not self._is_exist(self.directory) \
                or not self._is_exist(images_path) \
                or not self._is_exist(detections_path) \
                or not self._is_exist(ground_truths_path)\
                or not os.path.isfile(classes_txt_path):
            log.error("No such directory.")
            return False

        images = glob.glob(images_path + "/*")
        correct_files = []
        for image in images:
            detection = os.path.join(detections_path, f"{image.rsplit('/', 1)[1].rsplit('.', 1)[0]}.txt")
            if image.endswith((".jpeg", ".jpg", ".png", ".gif")) and os.path.isfile(detection):
                ground_truth = os.path.join(ground_truths_path, f"{image.rsplit('/', 1)[1].rsplit('.', 1)[0]}.txt")
                if os.path.isfile(ground_truth):
                    correct_files.extend([
                        {"type": "images", "path": image},
                        {"type": "detections", "path": detection},
                        {"type": "ground_truths", "path": ground_truth}
                    ])
                    continue
                ground_truth = os.path.join(ground_truths_path, f"{image.rsplit('/', 1)[1].rsplit('.', 1)[0]}.xml")
                if os.path.isfile(ground_truth):
                    correct_files.extend([
                        {"type": "images", "path": image},
                        {"type": "detections", "path": detection},
                        {"type": "ground_truths", "path": ground_truth}
                    ])

        if not correct_files:
            log.error("There are no files to upload.")
            return False

        correct_files.append({"type": "", "path": classes_txt_path})

        if self.upload_data(correct_files):
            return {
                "images_path": os.path.join(self.save_path, "images"),
                "ground_truths_path": os.path.join(self.save_path, "ground_truths"),
                "detections_path": os.path.join(self.save_path, "detections"),
                "classes_txt_path": os.path.join(self.save_path, "classes.txt")
            }



    def upload_data(self, data: list[dict]) -> bool:

        paths_count = len(data)
        progress_bar = tqdm(desc="Uploading...", total=paths_count)
        with concurrent.futures.ThreadPoolExecutor() as executor:
            upload_result = [executor.submit(self.__upload_file, file) for file in data]
            for _ in concurrent.futures.as_completed(upload_result):
                progress_bar.update(1)
        progress_bar.close()

        return True


    def __upload_file(self, file) -> None:

        url = f"{self.url}/api/v1/data/upload"
        files = {"file": open(file["path"], 'rb')}
        headers = {"token": self.token}
        data = {"data_path": os.path.join(self.save_path, file["type"])}
        response = requests.post(url, files=files, data=data, headers=headers)

        if response.status_code == 200:
            log.info(f'{file["path"]} is successfully uploaded.')
        else:
            log.error("Something went wrong.")


    def _is_exist(self, path: str) -> bool:
        if os.path.isdir(path):
            return True
        return False
